/* This work is part of OpenLDAP Software <http://www.openldap.org/>.
 *
 * Copyright 1998-2026 The OpenLDAP Foundation.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted only as authorized by the OpenLDAP
 * Public License.
 *
 * A copy of this license is available in the file LICENSE in the
 * top-level directory of the distribution or, alternatively, at
 * <http://www.OpenLDAP.org/license.html>.
 */

static const char copyright[] =
"Copyright 1998-2026 The OpenLDAP Foundation.  All rights reserved.\n"
"COPYING RESTRICTIONS APPLY\n";

static const char __Version[] =
"@(#) $OpenLDAP: dummyvalue 2.6.13 (" __DATE__ " " __TIME__ ") $\n"
"\topenldap\n";

